<?php

namespace InSegment\WorkersFramework\Services;

use Shavshukov\EventAppBase\ApplicationInterface;
use Shavshukov\EventAppBase\Interfaces\ModelInterface;
use React\Promise;

use InSegment\WorkersFramework\Interfaces\ConnectionListInterface;
use InSegment\WorkersFramework\Services\WorkerConsumer;
use InSegment\WorkersFramework\Models\State;
use InSegment\WorkersFramework\Queues\Worker;
use InSegment\WorkersFramework\Queues\ServiceMessenger;

class WorkerRunner
{
    /**
     * Application instance
     *
     * @var \Shavshukov\EventAppBase\ApplicationInterface
     */
    protected $app;
    
    /**
     * State instance
     *
     * @var \InSegment\WorkersFramework\Models\State
     */
    protected $state;
    
    /**
     * Worker to be run
     *
     * @var \InSegment\WorkersFramework\Queues\Worker
     */
    protected $worker;
    
    /**
     * Instance of queue driver connected to the channel
     *
     * @var \InSegment\WorkersFramework\Interfaces\ConnectionListInterface
     */
    protected $connections;
    
    /**
     * Consumer for state queue
     *
     * @var \InSegment\WorkersFramework\Services\StateConsumer
     */
    protected $stateConsumer;
    
    /**
     * Consumer for worker queue
     *
     * @var \InSegment\WorkersFramework\Services\WorkerConsumer
     */
    protected $workerConsumer;
    
    /**
     * Messenger for state changes between processes
     * 
     * @var \InSegment\WorkersFramework\Queues\ServiceMessenger
     */
    protected $serviceMessenger;
    
    /**
     * Constructor
     * 
     * @param \Shavshukov\EventAppBase\ApplicationInterface $app
     * @param \InSegment\WorkersFramework\Interfaces\ConnectionListInterface $connections
     * @param string $workerName
     * @throws \Exception
     */
    public function __construct(ApplicationInterface $app, ConnectionListInterface $connections, string $workerName)
    {
        $this->app = $app;
        $this->connections = $connections;
        $statePath = implode(DIRECTORY_SEPARATOR, [$this->app->getAppPath(), $app->config('state_file', 'storage/state.yaml')]);
        $this->state = new State($this->app->getFileReader(), $statePath);
        $this->worker = $this->findWorker($workerName);
        $this->stateConsumer = new StateConsumer($connections->getQueue($this->worker->getServiceQueueName()), $this->state);
        $this->stateConsumer->on('updates', function (ModelInterface $updates) use ($workerName) {
            $this->state->modify($updates->all());
            if ($updates->get('locked') || $updates->get("workers.{$workerName}.locked")) {
                $this->workerConsumer->deactivate()->then(function () {
                    $this->app->stop();
                });
            }
        });
        $this->workerConsumer = new WorkerConsumer($connections->getQueue($this->worker->getQueueName()), $app->config()->getJobList());
        $this->serviceMessenger = new ServiceMessenger($connections->getQueue($this->app->config('service_queue')));
    }
    
    /**
     * Run worker by name
     * 
     * @return \React\Promise\PromiseInterface
     */
    public function runWorker(): Promise\PromiseInterface
    {
        $this->state->load();
        // $this->state->modify(['timestamp' => microtime(true)]);
        
        if ($this->state->isWorkerLocked($this->worker->getName())) {
            return Promise\resolve($this->app->stop());
        } else {
            return Promise\all([
                $this->stateConsumer->activate(),
                $this->workerConsumer->activate(),
            ]);
        }
    }
    
    /**
     * Test worker by name and return a description of message count
     * 
     * @return \React\Promise\PromiseInterface<string>
     */
    public function testWorker(): Promise\PromiseInterface
    {
        $this->state->load();
        
        if ($this->state->get('locked')) {
            return Promise\resolve('Worker is locked');
        } else {
            return $this->worker->getDeclareStructure($this->connections)->declareForCount()->then(function ($messageCount) {
                return "Worker has {$messageCount} jobs pending";
            });
        }
    }
    
    /**
     * Find worker by name
     * 
     * @param string $workerName
     * @return \InSegment\WorkersFramework\Queues\Worker
     */
    protected function findWorker(string $workerName)
    {
        $nodeName = $this->app->config('node_name');
        $workerConfig = $this->app->config("workers.{$workerName}");
        if ($workerConfig) {
            return new Worker($workerName, $workerConfig, $nodeName);
        } else {
            echo "No such worker\n";
            exit(0);
        }
    }
}
